/*
 * Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/*
 * based in parts on libfaad, Copyright (C) 2003-2005 M. Bakker,
 * Ahead Software AG
 */

/** \file aac_afh.c para_server's aac audio format handler. */

#include <regex.h>
#include <mp4v2/mp4v2.h>

#include "para.h"
#include "error.h"
#include "portable_io.h"
#include "afh.h"
#include "string.h"
#include "aac.h"
#include "fd.h"

static int aac_find_stsz(char *buf, size_t buflen, off_t *skip)
{
	int i;

	for (i = 0; i + 16 < buflen; i++) {
		char *p = buf + i;
		unsigned sample_count, sample_size;

		if (p[0] != 's' || p[1] != 't' || p[2] != 's' || p[3] != 'z')
			continue;
		PARA_DEBUG_LOG("found stsz@%d\n", i);
		i += 8;
		sample_size = read_u32_be(buf + i);
		PARA_DEBUG_LOG("sample size: %u\n", sample_size);
		i += 4;
		sample_count = read_u32_be(buf + i);
		i += 4;
		PARA_DEBUG_LOG("sample count: %u\n", sample_count);
		*skip = i;
		return sample_count;
	}
	return -E_STSZ;
}

static int atom_cmp(const char *buf1, const char *buf2)
{
	return memcmp(buf1, buf2, 4)? 1 : 0;
}

static int read_atom_header(char *buf, uint64_t *subsize, char type[5])
{
	uint64_t size = read_u32_be(buf);

	memcpy(type, buf + 4, 4);
	type[4] = '\0';

	PARA_DEBUG_LOG("size: %llu, type: %s\n", (long long unsigned)size, type);
	if (size != 1) {
		*subsize = size;
		return 8;
	}
	buf += 4;
	size = 0;
	size = read_u64_be(buf);
	*subsize = size;
	return 16;
}

static char *get_tag(char *p, int size)
{
	char *buf;

	assert(size > 0);
	buf = para_malloc(size + 1);

	memcpy(buf, p, size);
	buf[size] = '\0';
	PARA_DEBUG_LOG("size: %d: %s\n", size, buf);
	return buf;
}

static void read_tags(char *buf, size_t buflen, struct afh_info *afhi)
{
	char *p = buf;

	while (p + 32 < buf + buflen) {
		char *q, type1[5], type2[5];
		uint64_t size1, size2;
		int ret, ret2;

		ret = read_atom_header(p, &size1, type1);
		ret2 = read_atom_header(p + ret, &size2, type2);

		if (size2 <= 16 || atom_cmp(type2, "data")) {
			p += size1;
			continue;
		}
		size2 -= 16;
		q = p + ret + ret2 + 8;
		if (q + size2 > buf + buflen)
			break;
		if (!atom_cmp(type1, "\xa9" "ART"))
			afhi->tags.artist = get_tag(q, size2);
		else if (!atom_cmp(type1, "\xa9" "alb"))
			afhi->tags.album = get_tag(q, size2);
		else if (!atom_cmp(type1, "\xa9" "nam"))
			afhi->tags.title = get_tag(q, size2);
		else if (!atom_cmp(type1, "\xa9" "cmt"))
			afhi->tags.comment = get_tag(q, size2);
		else if (!atom_cmp(type1, "\xa9" "day"))
			afhi->tags.year = get_tag(q, size2);
		p += size1;
	}
}

static void read_meta(char *buf, size_t buflen, struct afh_info *afhi)
{
	char *p = buf;

	while (p + 4 < buf + buflen) {

		if (p[0] != 'i' || p[1] != 'l' || p[2] != 's' || p[3] != 't') {
			p++;
			continue;
		}
		p += 4;
		return read_tags(p, buflen - (p - buf), afhi);
	}
}

static void aac_get_taginfo(char *buf, size_t buflen, struct afh_info *afhi)
{
	int i;
	uint64_t subsize;
	char type[5];

	for (i = 0; i + 24 < buflen; i++) {
		char *p = buf + i;
		if (p[0] != 'm' || p[1] != 'e' || p[2] != 't' || p[3] != 'a')
			continue;
		PARA_INFO_LOG("found metadata at offset %d\n", i);
		i += 8;
		p = buf + i;
		i += read_atom_header(p, &subsize, type);
		p = buf + i;
		return read_meta(p, buflen - i, afhi);
	}
	PARA_INFO_LOG("no meta data\n");
}

static ssize_t aac_compute_chunk_table(struct afh_info *afhi,
		char *map, size_t numbytes)
{
	int ret, i;
	size_t sum = 0;
	off_t skip;

	ret = aac_find_stsz(map, numbytes, &skip);
	if (ret < 0)
		return ret;
	afhi->chunks_total = ret;
	PARA_DEBUG_LOG("sz table has %" PRIu32 " entries\n", afhi->chunks_total);
	afhi->chunk_table = para_malloc((afhi->chunks_total + 1) * sizeof(size_t));
	for (i = 1; i <= afhi->chunks_total; i++) {
		if (skip + 4 > numbytes)
			break;
		sum += read_u32_be(map + skip);
		afhi->chunk_table[i] = sum;
		skip += 4;
//		if (i < 10 || i + 10 > afhi->chunks_total)
//			PARA_DEBUG_LOG("offset #%d: %zu\n", i, afhi->chunk_table[i]);
	}
	return skip;
}

static int aac_set_chunk_tv(struct afh_info *afhi,
		mp4AudioSpecificConfig *mp4ASC, uint32_t *seconds)
{
	float tmp = mp4ASC->sbr_present_flag == 1? 2047 : 1023;
	struct timeval total;
	long unsigned ms;

	if (!mp4ASC->samplingFrequency)
		return -E_MP4ASC;
	ms = 1000.0 * afhi->chunks_total * tmp / mp4ASC->samplingFrequency;
	ms2tv(ms, &total);
	tv_divide(afhi->chunks_total, &total, &afhi->chunk_tv);
	PARA_INFO_LOG("%luHz, %lus (%" PRIu32 " x %lums)\n",
		mp4ASC->samplingFrequency, ms / 1000,
		afhi->chunks_total, tv2ms(&afhi->chunk_tv));
	if (ms < 1000)
		return -E_MP4ASC;
	*seconds = ms / 1000;
	return 1;
}

/*
 * Init m4a file and write some tech data to given pointers.
 */
static int aac_get_file_info(char *map, size_t numbytes, __a_unused int fd,
		struct afh_info *afhi)
{
	int i;
	size_t skip;
	ssize_t ret;
	unsigned long rate = 0, decoder_len;
	unsigned char channels = 0;
	mp4AudioSpecificConfig mp4ASC;
	NeAACDecHandle handle = NULL;

	ret = aac_find_esds(map, numbytes, &skip, &decoder_len);
	if (ret < 0)
		goto out;
	aac_get_taginfo(map, numbytes, afhi);
	handle = aac_open();
	ret = -E_AAC_AFH_INIT;
	if (NeAACDecInit(handle, (unsigned char *)map + skip, decoder_len,
			&rate, &channels))
		goto out;
	if (!channels)
		goto out;
	PARA_DEBUG_LOG("rate: %lu, channels: %d\n", rate, channels);
	ret = -E_MP4ASC;
	if (NeAACDecAudioSpecificConfig((unsigned char *)map + skip,
			numbytes - skip, &mp4ASC))
		goto out;
	if (!mp4ASC.samplingFrequency)
		goto out;
	ret = aac_compute_chunk_table(afhi, map, numbytes);
	if (ret < 0)
		goto out;
	skip = ret;
	ret = aac_set_chunk_tv(afhi, &mp4ASC, &afhi->seconds_total);
	if (ret < 0)
		goto out;
	ret = aac_find_entry_point(map + skip, numbytes - skip, &skip);
	if (ret < 0)
		goto out;
	afhi->chunk_table[0] = ret;
	for (i = 1; i<= afhi->chunks_total; i++)
		afhi->chunk_table[i] += ret;
	afhi->channels = channels;
	afhi->frequency = rate;
	ret = (afhi->chunk_table[afhi->chunks_total] - afhi->chunk_table[0]) * 8; /* bits */
	ret += (channels * afhi->seconds_total * 500); /* avoid rounding error */
	afhi->bitrate = ret / (channels * afhi->seconds_total * 1000);
	ret = 1;
out:
	if (handle)
		NeAACDecClose(handle);
	return ret;
}

static int aac_rewrite_tags(const char *map, size_t mapsize,
		struct taginfo *tags, int fd, const char *filename)
{
	MP4FileHandle h;
	const MP4Tags *mdata;
	int ret = write_all(fd, map, mapsize);

	if (ret < 0)
		return ret;
	lseek(fd, 0, SEEK_SET);
	h = MP4Modify(filename, 0);
	if (!h) {
		PARA_ERROR_LOG("MP4Modify() failed, fd = %d\n", fd);
		return -E_MP4V2;
	}
	mdata = MP4TagsAlloc();
	assert(mdata);
	if (!MP4TagsFetch(mdata, h)) {
		PARA_ERROR_LOG("MP4Tags_Fetch() failed\n");
		ret = -E_MP4V2;
		goto close;
	}

	if (!MP4TagsSetAlbum(mdata, tags->album)) {
		PARA_ERROR_LOG("Could not set album\n");
		ret = -E_MP4V2;
		goto tags_free;
	}
	if (!MP4TagsSetArtist(mdata, tags->artist)) {
		PARA_ERROR_LOG("Could not set album\n");
		ret = -E_MP4V2;
		goto tags_free;
	}
	if (!MP4TagsSetComments(mdata, tags->comment)) {
		PARA_ERROR_LOG("Could not set comment\n");
		ret = -E_MP4V2;
		goto tags_free;
	}
	if (!MP4TagsSetName(mdata, tags->title)) {
		PARA_ERROR_LOG("Could not set title\n");
		ret = -E_MP4V2;
		goto tags_free;
	}
	if (!MP4TagsSetReleaseDate(mdata, tags->year)) {
		PARA_ERROR_LOG("Could not set release date\n");
		ret = -E_MP4V2;
		goto tags_free;
	}

	if (!MP4TagsStore(mdata, h)) {
		PARA_ERROR_LOG("Could not store tags\n");
		ret = -E_MP4V2;
		goto tags_free;
	}
	ret = 1;
tags_free:
	MP4TagsFree(mdata);
close:
	MP4Close(h, 0);
	return ret;
}

static const char * const aac_suffixes[] = {"m4a", "mp4", NULL};
/**
 * the init function of the aac audio format handler
 *
 * \param afh pointer to the struct to initialize
 */
void aac_afh_init(struct audio_format_handler *afh)
{
	afh->get_file_info = aac_get_file_info,
	afh->suffixes = aac_suffixes;
	afh->rewrite_tags = aac_rewrite_tags;
}
