/*
 * Copyright (C) 1998 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file gui.c Curses-based interface for paraslash. */

#include <regex.h>
#include <signal.h>
#include <sys/types.h>
#include <curses.h>
#include <locale.h>
#include <sys/time.h>
#include <lopsub.h>

#include "gui.lsg.h"
#include "para.h"
#include "gui.h"
#include "string.h"
#include "ringbuffer.h"
#include "fd.h"
#include "error.h"
#include "list.h"
#include "sched.h"
#include "signal.h"
#include "version.h"

/** Array of error strings. */
DEFINE_PARA_ERRLIST;

static struct lls_parse_result *cmdline_lpr, *lpr;

#define CMD_PTR (lls_cmd(0, gui_suite))
#define OPT_RESULT(_name) (lls_opt_result(LSG_GUI_PARA_GUI_OPT_ ## _name, lpr))
#define OPT_GIVEN(_name) (lls_opt_given(OPT_RESULT(_name)))
#define OPT_STRING_VAL(_name) (lls_string_val(0, OPT_RESULT(_name)))
#define OPT_UINT32_VAL(_name) (lls_uint32_val(0, OPT_RESULT(_name)))
#define FOR_EACH_KEY_MAP(_i) for (_i = 0; _i < OPT_GIVEN(KEY_MAP); _i++)

static char *stat_content[NUM_STAT_ITEMS];

static struct gui_window {
	WINDOW *win;
	bool needs_update;
} top, bot, sb, in, sep;

/** How many lines of output to remember. */
#define RINGBUFFER_SIZE 512

struct rb_entry {
	char *msg;
	size_t len;
	int color;
};
static struct ringbuffer *bot_win_rb;

static unsigned scroll_position;

static pid_t exec_pid;

static int exec_fds[2] = {-1, -1};
static int loglevel;

/** Type of the process currently being executed. */
enum exec_status {
	EXEC_IDLE, /**< No process running. */
	EXEC_DCMD, /**< para or display process running. */
	EXEC_XCMD, /**< External process running. */
};

/**
 * Codes for various colors.
 *
 * Each status item has its own color pair. The ones defined here start at a
 * higher number so that they do not overlap with these.
 */
enum gui_color_pair {
	COLOR_STATUSBAR = NUM_STAT_ITEMS + 1,
	COLOR_COMMAND,
	COLOR_OUTPUT,
	COLOR_MSG,
	COLOR_ERRMSG,
	COLOR_SEPARATOR,
	COLOR_TOP,
	COLOR_BOT,
};

struct gui_command {
	const char *key;
	const char *name;
	const char *description;
	void (*handler)(void);
};

static struct gui_theme theme;

#define GUI_COMMANDS \
	GUI_COMMAND(help, "?", "print help") \
	GUI_COMMAND(enlarge_top_win, "+", "enlarge the top window") \
	GUI_COMMAND(shrink_top_win, "-", "shrink the top window") \
	GUI_COMMAND(reread_conf, "r", "reread configuration file") \
	GUI_COMMAND(quit, "q", "exit para_gui") \
	GUI_COMMAND(refresh, "^L", "redraw the screen") \
	GUI_COMMAND(next_theme, ".", "switch to next theme") \
	GUI_COMMAND(prev_theme, ",", "switch to previous theme") \
	GUI_COMMAND(ll_incr, ">", "increase loglevel (decreases verbosity)") \
	GUI_COMMAND(ll_decr, "<", "decrease loglevel (increases verbosity)") \
	GUI_COMMAND(version, "V", "show the para_gui version") \
	GUI_COMMAND(scroll_up, "<up>", "scroll up one line") \
	GUI_COMMAND(scroll_down, "<down>", "scroll_down") \
	GUI_COMMAND(page_up, "<ppage>", "scroll up one page") \
	GUI_COMMAND(page_down, "<npage>", "scroll down one page") \
	GUI_COMMAND(scroll_top, "<home>", "scroll to top of buffer") \
	GUI_COMMAND(cancel_scroll, "<end>", "deactivate scroll mode") \

/* declare command handlers */
#define GUI_COMMAND(_c, _k, _d) \
	static void com_ ## _c(void);
GUI_COMMANDS

#undef GUI_COMMAND

/* define command array */
#define GUI_COMMAND(_c, _k, _d) \
	{ \
		.key = _k, \
		.name = #_c, \
		.description = _d, \
		.handler = com_ ## _c \
	},

static struct gui_command command_list[] = {GUI_COMMANDS {.name = NULL}};

struct input_task {
	struct task *task;
};

struct status_task {
	struct task *task;
	pid_t pid;
	char *buf;
	int bufsize, loaded;
	struct timeval next_exec;
	int fd;
};

/** Stdout/stderr of the executing process is read in chunks of this size. */
#define COMMAND_BUF_SIZE 32768

struct exec_task {
	struct task *task;
	char command_buf[2][COMMAND_BUF_SIZE]; /* stdout/stderr of command */
	int cbo[2]; /* command buf offsets */
	unsigned flags[2]; /* passed to for_each_line() */
};

static int find_cmd_byname(const char *name)
{
	int i;

	for (i = 0; command_list[i].handler; i++)
		if (!strcmp(command_list[i].name, name))
			return i;
	return -1;
}

/*
 * Even though ncurses provides getmaxx and getmaxy, these functions/macros are
 * not described in the XSI Curses standard.
 */
static int get_num_lines(struct gui_window *w)
{
	int lines;
	__a_unused int cols; /* avoid "set but not used" warnings */

	getmaxyx(w->win, lines, cols);
	return lines;
}

static int get_num_cols(struct gui_window *w)
{
	__a_unused int lines; /* avoid "set but not used" warnings */
	int cols;

	getmaxyx(w->win, lines, cols);
	return cols;
}

/** Number of lines of the window are occupied by an output line. */
#define NUM_LINES(len) (1 + (len) / get_num_cols(&bot))

/* isendwin() returns false before initscr() was called */
static bool curses_active(void)
{
	return top.win && !isendwin();
}

/* taken from mutt */
static char *km_keyname(int c)
{
	static char buf[10];

	if (c == KEY_UP) {
		sprintf(buf, "<up>");
		return buf;
	}
	if (c == KEY_DOWN) {
		sprintf(buf, "<down>");
		return buf;
	}
	if (c == KEY_LEFT) {
		sprintf(buf, "<left>");
		return buf;
	}
	if (c == KEY_RIGHT) {
		sprintf(buf, "<right>");
		return buf;
	}
	if (c == KEY_NPAGE) {
		sprintf(buf, "<npage>");
		return buf;
	}
	if (c == KEY_PPAGE) {
		sprintf(buf, "<ppage>");
		return buf;
	}
	if (c == KEY_HOME) {
		sprintf(buf, "<home>");
		return buf;
	}
	if (c == KEY_END) {
		sprintf(buf, "<end>");
		return buf;
	}
	if (c < 256 && c > -128 && iscntrl((unsigned char) c)) {
		if (c < 0)
			c += 256;
		if (c < 128) {
			buf[0] = '^';
			buf[1] = (c + '@') & 0x7f;
			buf[2] = 0;
		} else
			snprintf(buf, sizeof(buf), "\\%d%d%d", c >> 6,
				(c >> 3) & 7, c & 7);
	} else if (c >= KEY_F0 && c < KEY_F(256))
		sprintf(buf, "<F%d>", c - KEY_F0);
	else if (isprint(c))
		snprintf(buf, sizeof(buf), "%c", (unsigned char) c);
	else
		snprintf(buf, sizeof(buf), "\\x%hx", (unsigned short) c);
	return buf;
}

/* Print given number of spaces to curses window. */
static void add_spaces(WINDOW* win, unsigned int num)
{
	const char space[] = "                                ";
	const unsigned sz = sizeof(space) - 1; /* number of spaces */

	while (num >= sz)  {
		waddstr(win, space);
		num -= sz;
	}
	if (num > 0) {
		assert(num < sz);
		waddstr(win, space + sz - num);
	}
}

/*
 * print aligned string to curses window. This function always prints
 * exactly len chars.
 */
static int align_str(WINDOW* win, const char *str, unsigned int len,
		unsigned int align)
{
	int ret, num; /* of spaces */
	size_t width;
	char *sstr; /* sanitized string */

	if (!win || !str)
		return 0;
	ret = sanitize_str(str, len, &sstr, &width);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		width = 0;
		sstr = para_strdup(NULL);
	}
	assert(width <= len);
	num = len - width;
	if (align == LEFT) {
		waddstr(win, sstr);
		add_spaces(win, num);
	} else if (align == RIGHT) {
		add_spaces(win, num);
		waddstr(win, sstr);
	} else {
		add_spaces(win, num / 2);
		waddstr(win, sstr);
		add_spaces(win, num - num / 2);
	}
	free(sstr);
	return 1;
}

static void refresh_window(struct gui_window *gw)
{
	gw->needs_update = true;
}

static bool window_update_needed(void)
{
	return top.needs_update || bot.needs_update || sb.needs_update ||
		in.needs_update || sep.needs_update;
}

__printf_2_3 static void print_in_bar(int color, const char *fmt,...)
{
	char *msg;
	va_list ap;

	if (!curses_active())
		return;
	wattron(in.win, COLOR_PAIR(color));
	va_start(ap, fmt);
	xvasprintf(&msg, fmt, ap);
	va_end(ap);
	wmove(in.win, 0, 0);
	align_str(in.win, msg, get_num_cols(&in), LEFT);
	free(msg);
	refresh_window(&in);
}

static void print_status_bar(void)
{
	char *tmp;

	tmp = para_strdup("para_gui " PACKAGE_VERSION " (hit ? for help)");
	wmove(sb.win, 0, 0);
	align_str(sb.win, tmp, get_num_cols(&sb), CENTER);
	free(tmp);
}

/*
 * get the number of the oldest rbe that is (partially) visible. On return,
 * lines contains the sum of the number of lines of all visible entries. If the
 * first one is only partially visible, lines is greater than bot.lines.
 */
static int first_visible_rbe(unsigned *lines)
{
	int i, bot_lines = get_num_lines(&bot);

	*lines = 0;
	for (i = scroll_position; i < RINGBUFFER_SIZE; i++) {
		struct rb_entry *rbe = ringbuffer_get(bot_win_rb, i);
		int rbe_lines;
		if (!rbe)
			return i - 1;
		rbe_lines = NUM_LINES(rbe->len);
		if (rbe_lines > bot_lines)
			return -1;
		*lines += rbe_lines;
		if (*lines >= bot_lines)
			return i;
	}
	return RINGBUFFER_SIZE - 1;
}

/*
returns number of first visible rbe, *lines is the number of lines drawn.
 */
static int draw_top_rbe(unsigned *lines)
{
	int bot_cols, bot_lines, ret, fvr = first_visible_rbe(lines);
	struct rb_entry *rbe;
	size_t bytes_to_skip, cells_to_skip, width;

	if (fvr < 0)
		return -1;
	wmove(bot.win, 0, 0);
	rbe = ringbuffer_get(bot_win_rb, fvr);
	if (!rbe)
		return -1;
	getmaxyx(bot.win, bot_lines, bot_cols);
	if (*lines > bot_lines) {
		/* rbe is partially visible multi-line */
		cells_to_skip = (*lines - bot_lines) * bot_cols;
		ret = skip_cells(rbe->msg, cells_to_skip, &bytes_to_skip);
		if (ret < 0)
			return ret;
		ret = strwidth(rbe->msg + bytes_to_skip, &width);
		if (ret < 0)
			return ret;
	} else {
		bytes_to_skip = 0;
		width = rbe->len;
	}
	wattron(bot.win, COLOR_PAIR(rbe->color));
	waddstr(bot.win, rbe->msg + bytes_to_skip);
	*lines = NUM_LINES(width);
	return fvr;
}

static void redraw_bot_win(void)
{
	unsigned lines;
	int i, bot_lines = get_num_lines(&bot);

	wmove(bot.win, 0, 0);
	wclear(bot.win);
	i = draw_top_rbe(&lines);
	if (i <= 0)
		goto out;
	while (i > 0 && lines < bot_lines) {
		struct rb_entry *rbe = ringbuffer_get(bot_win_rb, --i);
		if (!rbe) {
			lines++;
			waddstr(bot.win, "\n");
			continue;
		}
		lines += NUM_LINES(rbe->len);
		wattron(bot.win, COLOR_PAIR(rbe->color));
		waddstr(bot.win, "\n");
		waddstr(bot.win, rbe->msg);
	}
out:
	refresh_window(&bot);
}

static void rb_add_entry(int color, char *msg)
{
	struct rb_entry *old, *new;
	int x, y;
	size_t len;

	if (strwidth(msg, &len) < 0)
		return;
	new = para_malloc(sizeof(struct rb_entry));
	new->color = color;
	new->len = len;
	new->msg = msg;
	old = ringbuffer_add(bot_win_rb, new);
	if (old) {
		free(old->msg);
		free(old);
	}
	if (scroll_position) {
		/* discard current scrolling, like xterm does */
		scroll_position = 0;
		redraw_bot_win();
		return;
	}
	wattron(bot.win, COLOR_PAIR(color));
	getyx(bot.win, y, x);
	if (y || x)
		waddstr(bot.win, "\n");
	waddstr(bot.win, msg);
}

/*
 * print formated output to bot win and refresh
 */
__printf_2_3 static void outputf(int color, const char* fmt,...)
{
	char *msg;
	va_list ap;

	if (!curses_active())
		return;
	va_start(ap, fmt);
	xvasprintf(&msg, fmt, ap);
	va_end(ap);
	rb_add_entry(color, msg);
	refresh_window(&bot);
}

static int add_output_line(char *line, void *data)
{
	int color = *(int *)data? COLOR_ERRMSG : COLOR_OUTPUT;

	if (!curses_active())
		return 1;
	rb_add_entry(color, para_strdup(line));
	return 1;
}

static __printf_2_3 void curses_log(int ll, const char *fmt,...)
{
	va_list ap;

	if (ll < loglevel)
		return;
	va_start(ap, fmt);
	if (curses_active()) {
		int color = ll <= LL_NOTICE? COLOR_MSG : COLOR_ERRMSG;
		char *msg;
		unsigned bytes = xvasprintf(&msg, fmt, ap);
		if (bytes > 0 && msg[bytes - 1] == '\n')
			msg[bytes - 1] = '\0'; /* cut trailing newline */
		rb_add_entry(color, msg);
		refresh_window(&bot);
	} else if (exec_pid <= 0) /* no external command running */
		vfprintf(stderr, fmt, ap);
	va_end(ap);
}
/** The log function of para_gui, always set to curses_log(). */
__printf_2_3 void (*para_log)(int, const char*, ...) = curses_log;

/* Call endwin() to reset the terminal into non-visual mode. */
static void shutdown_curses(void)
{
	/*
	 * If para_gui received a terminating signal in external mode, the
	 * terminal can be in an unusable state at this point because the child
	 * process might not have caught the signal. In this case endwin() has
	 * already been called and must not be called again. So we first return
	 * to program mode, then call endwin().
	 */
	if (!curses_active())
		reset_prog_mode();
	endwin();
}

/* disable curses, print a message, kill running processes and exit */
__noreturn __printf_2_3 static void die(int exit_code, const char* fmt, ...)
{
	va_list argp;

	/* Kill every process in our process group. */
	para_sigaction(SIGTERM, SIG_IGN);
	kill(0, SIGTERM);
	/* Wait up to two seconds for child processes to die. */
	alarm(2);
	while (waitpid(0, NULL, 0) >= 0)
		; /* nothing */
	alarm(0);
	/* mousemask() exists only in ncurses */
#ifdef NCURSES_MOUSE_VERSION
	mousemask(~(mmask_t)0, NULL); /* Avoid bad terminal state with xterm. */
#endif
	shutdown_curses();
	va_start(argp, fmt);
	vfprintf(stderr, fmt, argp);
	va_end(argp);
	exit(exit_code);
}

/*
 * Print stat item #i to curses window
 */
static void print_stat_item(int i)
{
	char *tmp;
	struct stat_item_data d = theme.data[i];
	char *c = stat_content[i];
	int top_lines = get_num_lines(&top);

	if (!curses_active() || !d.len || !c)
		return;
	tmp = make_message("%s%s%s", d.prefix, c, d.postfix);
	wmove(top.win, d.y * top_lines / 100, d.x * COLS / 100);
	wattron(top.win, COLOR_PAIR(i + 1));
	align_str(top.win, tmp, d.len * COLS / 100, d.align);
	free(tmp);
	refresh_window(&top);
}

static int update_item(int item_num, char *buf)
{
	char **c = stat_content + item_num;

	free(*c);
	if (buf && buf[0])
		goto dup;
	switch (item_num) {
	case SI_ARTIST:
		*c = para_strdup("(artist tag not set)");
		goto print;
	case SI_TITLE:
		*c = para_strdup("(title tag not set)");
		goto print;
	case SI_YEAR:
		*c = para_strdup("????");
		goto print;
	case SI_ALBUM:
		*c = para_strdup("(album tag not set)");
		goto print;
	case SI_COMMENT:
		*c = para_strdup("(comment tag not set)");
		goto print;
	}
dup:
	*c = para_strdup(buf);
print:
	print_stat_item(item_num);
	return 1;
}

static void print_all_items(void)
{
	int i;

	if (!curses_active())
		return;
	FOR_EACH_STATUS_ITEM(i)
		print_stat_item(i);
}

static void clear_all_items(void)
{
	int i;

	FOR_EACH_STATUS_ITEM(i) {
		free(stat_content[i]);
		stat_content[i] = para_strdup("");
	}
}

static void status_pre_select(struct sched *s, void *context)
{
	struct status_task *st = context;

	if (st->fd >= 0)
		para_fd_set(st->fd, &s->rfds, &s->max_fileno);
	if (task_get_notification(st->task) < 0)
		return sched_min_delay(s);
	if (st->fd < 0)
		sched_request_barrier_or_min_delay(&st->next_exec, s);
}

static int status_post_select(struct sched *s, void *context)
{
	struct status_task *st = context;
	size_t sz;
	int ret, ret2;

	ret = task_get_notification(st->task);
	if (ret == -E_GUI_SIGCHLD && st->pid > 0) {
		int exit_status;
		if (waitpid(st->pid, &exit_status, WNOHANG) == st->pid) {
			st->pid = 0;
			PARA_ERROR_LOG("stat command exit status: %d",
				exit_status);
		}
	}
	if (st->fd < 0) {
		int fds[3] = {0, 1, 0};
		if (st->pid > 0)
			return 0;
		/* Avoid busy loop */
		if (tv_diff(&st->next_exec, now, NULL) > 0)
			return 0;
		st->next_exec.tv_sec = now->tv_sec + 2;
		ret = para_exec_cmdline_pid(&st->pid,
			OPT_STRING_VAL(STAT_CMD), fds);
		if (ret < 0)
			return 0;
		ret = mark_fd_nonblocking(fds[1]);
		if (ret < 0) {
			close(fds[1]);
			return 0;
		}
		st->fd = fds[1];
		return 0;
	}

	if (st->loaded >= st->bufsize) {
		if (st->bufsize > 1000 * 1000) {
			st->loaded = 0;
			return 0;
		}
		st->bufsize += st->bufsize + 1000;
		st->buf = para_realloc(st->buf, st->bufsize);
	}
	assert(st->loaded < st->bufsize);
	ret = read_nonblock(st->fd, st->buf + st->loaded,
		st->bufsize - st->loaded, &s->rfds, &sz);
	st->loaded += sz;
	ret2 = for_each_stat_item(st->buf, st->loaded, update_item);
	if (ret < 0 || ret2 < 0) {
		st->loaded = 0;
		PARA_NOTICE_LOG("closing stat pipe: %s\n",
			para_strerror(ret < 0? -ret : -ret2));
		close(st->fd);
		st->fd = -1;
		clear_all_items();
		free(stat_content[SI_BASENAME]);
		stat_content[SI_BASENAME] =
			para_strdup("stat command terminated!?");
		print_all_items();
		return 0;
	}
	sz = ret2; /* what is left */
	if (sz > 0 && sz < st->loaded)
		memmove(st->buf, st->buf + st->loaded - sz, sz);
	st->loaded = sz;
	return 0;
}

/*
 * init all windows
 */
static void init_wins(int top_lines)
{
	int top_y = 0, bot_y = top_lines + 1, sb_y = LINES - 2,
		in_y = LINES - 1, sep_y = top_lines;
	int bot_lines = LINES - top_lines - 3, sb_lines = 1, in_lines = 1,
		sep_lines = 1;

	assume_default_colors(theme.dflt.fg, theme.dflt.bg);
	if (top.win) {
		wresize(top.win, top_lines, COLS);
		mvwin(top.win, top_y, 0);

		wresize(sb.win, sb_lines, COLS);
		mvwin(sb.win, sb_y, 0);

		wresize(sep.win, sep_lines, COLS);
		mvwin(sep.win, sep_y, 0);

		wresize(bot.win, bot_lines, COLS);
		mvwin(bot.win, bot_y, 0);

		wresize(in.win, in_lines, COLS);
		mvwin(in.win, in_y, 0);
	} else {
		sep.win = newwin(sep_lines, COLS, sep_y, 0);
		top.win = newwin(top_lines, COLS, top_y, 0);
		bot.win = newwin(bot_lines, COLS, bot_y, 0);
		sb.win = newwin(sb_lines, COLS, sb_y, 0);
		in.win = newwin(in_lines, COLS, in_y, 0);
		if (!top.win || !bot.win || !sb.win || !in.win || !sep.win)
			die(EXIT_FAILURE, "Error: Cannot create curses windows\n");
		wclear(bot.win);
		wclear(sb.win);
		wclear(in.win);
		scrollok(bot.win, 1);
		wattron(sb.win, COLOR_PAIR(COLOR_STATUSBAR));
		wattron(sep.win, COLOR_PAIR(COLOR_SEPARATOR));
		wattron(bot.win, COLOR_PAIR(COLOR_BOT));
		wattron(top.win, COLOR_PAIR(COLOR_TOP));
		nodelay(top.win, 1);
		nodelay(bot.win, 1);
		nodelay(sb.win, 1);
		nodelay(in.win, 0);

		keypad(top.win, 1);
		keypad(bot.win, 1);
		keypad(sb.win, 1);
		keypad(in.win, 1);
	}
	wmove(sep.win, 0, 0);
	whline(sep.win, theme.sep_char, COLS);
	wclear(top.win);
	print_all_items();
	//wclear(bot.win);
	wnoutrefresh(top.win);
	wnoutrefresh(bot.win);
	print_status_bar();
	wnoutrefresh(sb.win);
	wnoutrefresh(in.win);
	wnoutrefresh(sep.win);
	doupdate();
}

static void init_pair_or_die(short pair, short f, short b)
{
	if (init_pair(pair, f, b) == ERR)
		die(EXIT_FAILURE, "fatal: init_pair() failed\n");
}

static void init_colors_or_die(void)
{
	int i;

	if (!has_colors())
		die(EXIT_FAILURE, "fatal: No color term\n");
	if (start_color() == ERR)
		die(EXIT_FAILURE, "fatal: failed to start colors\n");
	FOR_EACH_STATUS_ITEM(i)
		if (theme.data[i].len)
			init_pair_or_die(i + 1, theme.data[i].color.fg,
				theme.data[i].color.bg);
	init_pair_or_die(COLOR_STATUSBAR, theme.sb.fg, theme.sb.bg);
	init_pair_or_die(COLOR_COMMAND, theme.cmd.fg, theme.cmd.bg);
	init_pair_or_die(COLOR_OUTPUT, theme.output.fg, theme.output.bg);
	init_pair_or_die(COLOR_MSG, theme.msg.fg, theme.msg.bg);
	init_pair_or_die(COLOR_ERRMSG, theme.err_msg.fg, theme.err_msg.bg);
	init_pair_or_die(COLOR_SEPARATOR, theme.sep.fg, theme.sep.bg);
	init_pair_or_die(COLOR_TOP, theme.dflt.fg, theme.dflt.bg);
	init_pair_or_die(COLOR_BOT, theme.dflt.fg, theme.dflt.bg);
}

/* (Re-)initialize the curses library. */
static void init_curses(void)
{
	if (curses_active())
		return;
	if (refresh() == ERR) /* refresh is really needed */
		die(EXIT_FAILURE, "refresh() failed\n");
	if (LINES < theme.lines_min || COLS < theme.cols_min)
		die(EXIT_FAILURE, "Terminal (%dx%d) too small"
			" (need at least %dx%d)\n", COLS, LINES,
			theme.cols_min, theme.lines_min);
	curs_set(0); /* make cursor invisible, ignore errors */
	nonl(); /* do not NL->CR/NL on output, always returns OK */
	/* don't echo input */
	if (noecho() == ERR)
		die(EXIT_FAILURE, "fatal: noecho() failed\n");
	/* take input chars one at a time, no wait for \n */
	if (cbreak() == ERR)
		die(EXIT_FAILURE, "fatal: cbreak() failed\n");
	init_colors_or_die();
	clear(); /* ignore non-fatal errors */
	init_wins(theme.top_lines_default);
	// noecho(); /* don't echo input */
}

/*
 * This sucker modifies its first argument. *handler and *arg are
 * pointers to 0-terminated strings (inside line). Crap.
 */
static int split_key_map(char *line, char **handler, char **arg)
{
	if (!(*handler = strchr(line + 1, ':')))
		goto err_out;
	**handler = '\0';
	(*handler)++;
	if (!(*arg = strchr(*handler, ':')))
		goto err_out;
	**arg = '\0';
	(*arg)++;
	return 1;
err_out:
	return 0;
}

static void check_key_map_args_or_die(void)
{
	int i;
	char *tmp = NULL;
	const struct lls_opt_result *lor = OPT_RESULT(KEY_MAP);

	FOR_EACH_KEY_MAP(i) {
		char *handler, *arg;

		free(tmp);
		tmp = para_strdup(lls_string_val(i, lor));
		if (!split_key_map(tmp, &handler, &arg))
			break;
		if (strlen(handler) != 1)
			break;
		if (*handler != 'x' && *handler != 'd' && *handler != 'i'
				&& *handler != 'p')
			break;
		if (*handler != 'i')
			continue;
		if (find_cmd_byname(arg) < 0)
			break;
	}
	if (i != OPT_GIVEN(KEY_MAP))
		die(EXIT_FAILURE, "invalid key map: %s\n",
			lls_string_val(i, lor));
	free(tmp);
}

static void parse_config_file_or_die(bool reload)
{
	int ret;
	char *cf = NULL, *errctx = NULL;
	void *map;
	size_t sz;
	int cf_argc;
	char **cf_argv;
	struct lls_parse_result *cf_lpr, *merged_lpr;

	if (OPT_GIVEN(CONFIG_FILE))
		cf = para_strdup(OPT_STRING_VAL(CONFIG_FILE));
	else {
		char *home = para_homedir();
		cf = make_message("%s/.paraslash/gui.conf", home);
		free(home);
	}
	ret = mmap_full_file(cf, O_RDONLY, &map, &sz, NULL);
	if (ret < 0) {
		if (ret != -E_EMPTY && ret != -ERRNO_TO_PARA_ERROR(ENOENT))
			goto free_cf;
		if (ret == -ERRNO_TO_PARA_ERROR(ENOENT) && OPT_GIVEN(CONFIG_FILE))
			goto free_cf;
		ret = 0;
		lpr = cmdline_lpr;
		goto success;
	}
	ret = lls(lls_convert_config(map, sz, NULL, &cf_argv, &errctx));
	para_munmap(map, sz);
	if (ret < 0)
		goto free_cf;
	cf_argc = ret;
	ret = lls(lls_parse(cf_argc, cf_argv, CMD_PTR, &cf_lpr, &errctx));
	lls_free_argv(cf_argv);
	if (ret < 0)
		goto free_cf;
	if (reload) /* config file overrides command line */
		ret = lls(lls_merge(cf_lpr, cmdline_lpr, CMD_PTR, &merged_lpr,
			&errctx));
	else /* command line options overrride config file options */
		ret = lls(lls_merge(cmdline_lpr, cf_lpr, CMD_PTR, &merged_lpr,
			&errctx));
	lls_free_parse_result(cf_lpr, CMD_PTR);
	if (ret < 0)
		goto free_cf;
	if (lpr != cmdline_lpr)
		lls_free_parse_result(lpr, CMD_PTR);
	lpr = merged_lpr;
success:
	loglevel = OPT_UINT32_VAL(LOGLEVEL);
	check_key_map_args_or_die();
	theme_init(OPT_STRING_VAL(THEME), &theme);
free_cf:
	free(cf);
	if (ret < 0) {
		if (errctx)
			PARA_ERROR_LOG("%s\n", errctx);
		free(errctx);
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
}

/* reread configuration, terminate on errors */
static void reread_conf(void)
{
	/*
	 * If the reload of the config file fails, we are about to exit. In
	 * this case we print the error message to stderr rather than to the
	 * curses window.  So we have to shutdown curses first.
	 */
	shutdown_curses();
	parse_config_file_or_die(true);
	init_curses();
	print_in_bar(COLOR_MSG, "config file reloaded\n");
}

/*
 * React to various signal-related events
 */
static int signal_post_select(struct sched *s, __a_unused void *context)
{
	int ret = para_next_signal(&s->rfds);

	if (ret <= 0)
		return 0;
	switch (ret) {
	case SIGTERM:
		die(EXIT_FAILURE, "only the good die young (caught SIGTERM)\n");
	case SIGINT:
		return 1;
	case SIGUSR1:
		PARA_NOTICE_LOG("got SIGUSR1, rereading configuration\n");
		reread_conf();
		return 1;
	case SIGCHLD:
		task_notify_all(s, E_GUI_SIGCHLD);
		return 1;
	}
	return 1;
}

static enum exec_status exec_status(void)
{
	if (exec_fds[0] >= 0 || exec_fds[1] >= 0)
		return EXEC_DCMD;
	if (exec_pid > 0)
		return EXEC_XCMD;
	return EXEC_IDLE;
}

static void exec_pre_select(struct sched *s, void *context)
{
	struct exec_task *et = context;
	if (exec_fds[0] >= 0)
		para_fd_set(exec_fds[0], &s->rfds, &s->max_fileno);
	if (exec_fds[1] >= 0)
		para_fd_set(exec_fds[1], &s->rfds, &s->max_fileno);
	if (task_get_notification(et->task) < 0)
		sched_min_delay(s);
}

static int exec_post_select(struct sched *s, void *context)
{
	struct exec_task *ct = context;
	int i, ret;

	ret = task_get_notification(ct->task);
	if (ret == -E_GUI_SIGCHLD && exec_pid > 0) {
		int exit_status;
		if (waitpid(exec_pid, &exit_status, WNOHANG) == exec_pid) {
			exec_pid = 0;
			init_curses();
			PARA_INFO_LOG("command exit status: %d", exit_status);
			print_in_bar(COLOR_MSG, " ");
		}
	}
	for (i = 0; i < 2; i++) {
		size_t sz;
		if (exec_fds[i] < 0)
			continue;
		ret = read_nonblock(exec_fds[i],
			ct->command_buf[i] + ct->cbo[i],
			COMMAND_BUF_SIZE - 1 - ct->cbo[i], &s->rfds, &sz);
		ct->cbo[i] += sz;
		sz = ct->cbo[i];
		ct->cbo[i] = for_each_line(ct->flags[i], ct->command_buf[i],
			ct->cbo[i], add_output_line, &i);
		if (sz != ct->cbo[i]) { /* at least one line found */
			refresh_window(&bot);
			ct->flags[i] = 0;
		}
		if (ret < 0 || exec_pid == 0) {
			if (ret < 0)
				PARA_NOTICE_LOG("closing command fd %d: %s",
					i, para_strerror(-ret));
			close(exec_fds[i]);
			exec_fds[i] = -1;
			ct->flags[i] = 0;
			ct->cbo[i] = 0;
			if (exec_fds[!i] < 0) /* both fds closed */
				return 1;
		}
		if (ct->cbo[i] == COMMAND_BUF_SIZE - 1) {
			PARA_NOTICE_LOG("discarding overlong line");
			ct->cbo[i] = 0;
			ct->flags[i] = FELF_DISCARD_FIRST;
		}
	}
	return 0;
}

static void input_pre_select(struct sched *s, __a_unused void *context)
{
	if (exec_status() != EXEC_XCMD)
		para_fd_set(STDIN_FILENO, &s->rfds, &s->max_fileno);
	if (window_update_needed())
		sched_min_delay(s);
}

/* read from command pipe and print data to bot window */
static void exec_and_display(const char *file_and_args)
{
	int ret, fds[3] = {0, 1, 1};

	outputf(COLOR_COMMAND, "%s", file_and_args);
	ret = para_exec_cmdline_pid(&exec_pid, file_and_args, fds);
	if (ret < 0)
		return;
	ret = mark_fd_nonblocking(fds[1]);
	if (ret < 0)
		goto fail;
	ret = mark_fd_nonblocking(fds[2]);
	if (ret < 0)
		goto fail;
	exec_fds[0] = fds[1];
	exec_fds[1] = fds[2];
	print_in_bar(COLOR_MSG, "hit any key to abort\n");
	return;
fail:
	PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	close(exec_fds[0]);
	close(exec_fds[1]);
}

static void exec_para(const char *args)
{
	char *file_and_args;

	file_and_args = make_message(BINDIR "/para_client -- %s", args);
	exec_and_display(file_and_args);
	free(file_and_args);
}

/*
 * shutdown curses and stat pipe before executing external commands
 */
static void exec_external(char *file_and_args)
{
	int fds[3] = {-1, -1, -1};

	if (exec_pid)
		return;
	shutdown_curses();
	para_exec_cmdline_pid(&exec_pid, file_and_args, fds);
}

static void handle_command(int c)
{
	int i;
	const struct lls_opt_result *lor = OPT_RESULT(KEY_MAP);

	/* first check user-defined key bindings */
	FOR_EACH_KEY_MAP(i) {
		char *tmp, *handler, *arg;

		tmp = para_strdup(lls_string_val(i, lor));
		if (!split_key_map(tmp, &handler, &arg)) {
			free(tmp);
			return;
		}
		if (strcmp(tmp, km_keyname(c))) {
			free(tmp);
			continue;
		}
		if (*handler == 'd')
			exec_and_display(arg);
		else if (*handler == 'x')
			exec_external(arg);
		else if (*handler == 'p')
			exec_para(arg);
		else if (*handler == 'i') {
			int num = find_cmd_byname(arg);
			if (num >= 0)
				command_list[num].handler();
		}
		free(tmp);
		return;
	}
	/* not found, check internal key bindings */
	for (i = 0; command_list[i].handler; i++) {
		if (!strcmp(km_keyname(c), command_list[i].key)) {
			command_list[i].handler();
			return;
		}
	}
	print_in_bar(COLOR_ERRMSG, "key '%s' is not bound, press ? for help",
		km_keyname(c));
}

static int input_post_select(__a_unused struct sched *s, __a_unused void *context)
{
	int ret;
	enum exec_status exs = exec_status();

	if (exs == EXEC_XCMD)
		return 0;
	if (window_update_needed()) {
		if (top.needs_update)
			assert(wnoutrefresh(top.win) == OK);
		if (bot.needs_update)
			assert(wnoutrefresh(bot.win) == OK);
		if (sep.needs_update)
			assert(wnoutrefresh(sep.win) == OK);
		if (sb.needs_update)
			assert(wnoutrefresh(sb.win) == OK);
		if (in.needs_update)
			assert(wnoutrefresh(in.win) == OK);
		doupdate();
		top.needs_update = bot.needs_update = sb.needs_update =
			in.needs_update = sep.needs_update = false;
	}
	ret = wgetch(top.win);
	if (ret == ERR)
		return 0;
	if (ret == KEY_RESIZE) {
		if (curses_active()) {
			shutdown_curses();
			init_curses();
			redraw_bot_win();
		}
		return 0;
	}
	if (exs == EXEC_IDLE)
		handle_command(ret);
	else if (exec_pid > 0)
		kill(exec_pid, SIGTERM);
	return 0;
}

static void print_scroll_msg(void)
{
	unsigned lines_total, filled = ringbuffer_filled(bot_win_rb);
	int first_rbe = first_visible_rbe(&lines_total);

	print_in_bar(COLOR_MSG, "scrolled view: %u-%u/%u\n", filled - first_rbe,
		filled - scroll_position, ringbuffer_filled(bot_win_rb));
}

static void com_scroll_top(void)
{
	int i = RINGBUFFER_SIZE - 1, bot_lines = get_num_lines(&bot);
	unsigned lines = 0;

	while (i > 0 && !ringbuffer_get(bot_win_rb, i))
		i--;
	/* i is oldest entry */
	for (; lines < bot_lines && i >= 0; i--) {
		struct rb_entry *rbe = ringbuffer_get(bot_win_rb, i);
		if (!rbe)
			break;
		lines += NUM_LINES(rbe->len);
	}
	i++;
	if (lines > 0 && scroll_position != i) {
		scroll_position = i;
		redraw_bot_win();
		print_scroll_msg();
		return;
	}
	print_in_bar(COLOR_ERRMSG, "top of buffer is shown\n");
}

static void com_cancel_scroll(void)
{

	if (scroll_position == 0) {
		print_in_bar(COLOR_ERRMSG, "bottom of buffer is shown\n");
		return;
	}
	scroll_position = 0;
	redraw_bot_win();
}

static void com_page_down(void)
{
	unsigned lines = 0;
	int i = scroll_position, bot_lines = get_num_lines(&bot);

	while (lines < bot_lines && --i > 0) {
		struct rb_entry *rbe = ringbuffer_get(bot_win_rb, i);
		if (!rbe)
			break;
		lines += NUM_LINES(rbe->len);
	}
	if (lines) {
		scroll_position = i;
		redraw_bot_win();
		print_scroll_msg();
		return;
	}
	print_in_bar(COLOR_ERRMSG, "bottom of buffer is shown\n");
}

static void com_page_up(void)
{
	unsigned lines;
	int fvr = first_visible_rbe(&lines), bot_lines = get_num_lines(&bot);

	if (fvr < 0 || fvr + 1 >= ringbuffer_filled(bot_win_rb)) {
		print_in_bar(COLOR_ERRMSG, "top of buffer is shown\n");
		return;
	}
	scroll_position = fvr + 1;
	for (; scroll_position > 0; scroll_position--) {
		first_visible_rbe(&lines);
		if (lines == bot_lines)
			break;
	}
	redraw_bot_win();
	print_scroll_msg();
}

static void com_scroll_down(void)
{
	struct rb_entry *rbe;
	int rbe_lines, bot_lines = get_num_lines(&bot);

	if (!scroll_position) {
		print_in_bar(COLOR_ERRMSG, "bottom of buffer is shown\n");
		return;
	}
	scroll_position--;
	rbe = ringbuffer_get(bot_win_rb, scroll_position);
	rbe_lines = NUM_LINES(rbe->len);
	wscrl(bot.win, rbe_lines);
	wmove(bot.win, bot_lines - rbe_lines, 0);
	wattron(bot.win, COLOR_PAIR(rbe->color));
	waddstr(bot.win, rbe->msg);
	refresh_window(&bot);
	print_scroll_msg();
}

static void com_scroll_up(void)
{
	struct rb_entry *rbe = NULL;
	unsigned lines;
	int i, first_rbe, num_scroll;

	/* the entry that is going to vanish */
	rbe = ringbuffer_get(bot_win_rb, scroll_position);
	if (!rbe)
		goto err_out;
	num_scroll = NUM_LINES(rbe->len);
	first_rbe = first_visible_rbe(&lines);
	if (first_rbe < 0 || (first_rbe == ringbuffer_filled(bot_win_rb) - 1))
		goto err_out;
	scroll_position++;
	wscrl(bot.win, -num_scroll);
	i = draw_top_rbe(&lines);
	if (i < 0)
		goto err_out;
	while (i > 0 && lines < num_scroll) {
		int rbe_lines;
		rbe = ringbuffer_get(bot_win_rb, --i);
		if (!rbe)
			break;
		rbe_lines = NUM_LINES(rbe->len);
		lines += rbe_lines;
		wattron(bot.win, COLOR_PAIR(rbe->color));
		waddstr(bot.win, "\n");
		waddstr(bot.win, rbe->msg);
		if (!i)
			break;
		i--;
	}
	refresh_window(&bot);
	print_scroll_msg();
	return;
err_out:
	print_in_bar(COLOR_ERRMSG, "top of buffer is shown\n");
}

static void com_ll_decr(void)
{
	if (loglevel <= LL_DEBUG) {
		print_in_bar(COLOR_ERRMSG,
			"loglevel already at maximal verbosity\n");
		return;
	}
	loglevel--;
	print_in_bar(COLOR_MSG, "loglevel set to %d\n", loglevel);
}

static void com_ll_incr(void)
{
	if (loglevel >= LL_EMERG) {
		print_in_bar(COLOR_ERRMSG,
			"loglevel already at minimal verbosity\n");
		return;
	}
	loglevel++;
	print_in_bar(COLOR_MSG, "loglevel set to %d\n", loglevel);
}

static void com_reread_conf(void)
{
	reread_conf();
}

static void com_help(void)
{
	int i;
	const struct lls_opt_result *lor = OPT_RESULT(KEY_MAP);

	FOR_EACH_KEY_MAP(i) {
		char *handler, *arg, *tmp = para_strdup(lls_string_val(i, lor));
		const char *handler_text = "???", *desc = NULL;

		if (!split_key_map(tmp, &handler, &arg)) {
			free(tmp);
			return;
		}
		switch (*handler) {
			case 'i':
				handler_text = "internal";
				desc = command_list[find_cmd_byname(arg)].description;
				break;
			case 'x': handler_text = "external"; break;
			case 'd': handler_text = "display "; break;
			case 'p': handler_text = "para    "; break;
		}
		outputf(COLOR_MSG, "%s\t%s\t%s%s\t%s", tmp, handler_text, arg,
			strlen(arg) < 8? "\t" : "",
			desc? desc : "");
		free(tmp);
	}
	for (i = 0; command_list[i].handler; i++) {
		struct gui_command gc = command_list[i];

		outputf(COLOR_MSG, "%s\tinternal\t%s\t%s%s", gc.key, gc.name,
			strlen(gc.name) < 8? "\t" : "",
			gc.description);
	}
	print_in_bar(COLOR_MSG, "try \"para_gui -h\" or \"para_client help\" "
		"for more info");
}

static void com_shrink_top_win(void)
{
	int top_lines = get_num_lines(&top);

	if (top_lines <= theme.top_lines_min) {
		PARA_WARNING_LOG("can not decrease top window\n");
		return;
	}
	init_wins(top_lines - 1);
	print_in_bar(COLOR_MSG, "%s", "decreased top window");
}

static void com_enlarge_top_win(void)
{
	int top_lines = get_num_lines(&top), bot_lines = get_num_lines(&bot);

	if (bot_lines < 3) {
		PARA_WARNING_LOG("can not increase top window\n");
		return;
	}
	init_wins(top_lines + 1);
	print_in_bar(COLOR_MSG, "increased top window");
}

static void com_version(void)
{
	print_in_bar(COLOR_MSG, "%s", version_single_line("gui"));
}

__noreturn static void com_quit(void)
{
	die(EXIT_SUCCESS, "%s", "");
}

static void com_refresh(void)
{
	shutdown_curses();
	init_curses();
}

static void com_next_theme(void)
{
	theme_next(&theme);
	com_refresh();
}

static void com_prev_theme(void)
{
	theme_prev(&theme);
	com_refresh();
}

static int setup_tasks_and_schedule(void)
{
	int ret;
	struct exec_task exec_task = {.task = NULL};
	struct status_task status_task = {.fd = -1};
	struct input_task input_task = {.task = NULL};
	struct signal_task *signal_task;
	struct sched sched = {.default_timeout = {.tv_sec = 1}};

	exec_task.task = task_register(&(struct task_info) {
		.name = "exec",
		.pre_select = exec_pre_select,
		.post_select = exec_post_select,
		.context = &exec_task,
	}, &sched);

	status_task.task = task_register(&(struct task_info) {
		.name = "status",
		.pre_select = status_pre_select,
		.post_select = status_post_select,
		.context = &status_task,
	}, &sched);

	input_task.task = task_register(&(struct task_info) {
		.name = "input",
		.pre_select = input_pre_select,
		.post_select = input_post_select,
		.context = &input_task,
	}, &sched);

	signal_task = signal_init_or_die();
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGCHLD);
	para_install_sighandler(SIGUSR1);
	signal_task->task = task_register(&(struct task_info) {
		.name = "signal",
		.pre_select = signal_pre_select,
		.post_select = signal_post_select,
		.context = signal_task,
	}, &sched);
	ret = schedule(&sched);
	sched_shutdown(&sched);
	signal_shutdown(signal_task);
	return ret;
}

static void handle_help_flags(void)
{
	char *help;

	if (OPT_GIVEN(DETAILED_HELP))
		help = lls_long_help(CMD_PTR);
	else if (OPT_GIVEN(HELP))
		help = lls_short_help(CMD_PTR);
	else
		return;
	printf("%s\n", help);
	free(help);
	exit(EXIT_SUCCESS);
}

/**
 * The main function of para_gui.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 *
 * After initialization para_gui registers the following tasks to the paraslash
 * scheduler: status, exec, signal, input.
 *
 * The status task executes the para_audioc stat command to obtain the status
 * of para_server and para_audiod, and displays this information in the top
 * window of para_gui.
 *
 * The exec task is responsible for printing the output of the currently
 * running executable to the bottom window.
 *
 * The signal task performs various actions according to signals received. For
 * example, it reloads the configuration file on SIGUSR1, and it shuts down the
 * curses system on SIGTERM to restore the terminal settings before exit.
 *
 * The input task reads single key strokes from stdin. For each key pressed, it
 * executes the command handler associated with this key.
 *
 * \return \p EXIT_SUCCESS or \p EXIT_FAILURE.
 */
int main(int argc, char *argv[])
{
	int ret;
	char *errctx;

	ret = lls(lls_parse(argc, argv, CMD_PTR, &cmdline_lpr, &errctx));
	if (ret < 0)
		goto out;
	lpr = cmdline_lpr;
	loglevel = OPT_UINT32_VAL(LOGLEVEL);
	version_handle_flag("gui", OPT_GIVEN(VERSION));
	handle_help_flags();
	parse_config_file_or_die(false);
	bot_win_rb = ringbuffer_new(RINGBUFFER_SIZE);
	setlocale(LC_CTYPE, "");
	initscr(); /* needed only once, always successful */
	init_curses();
	ret = setup_tasks_and_schedule();
out:
	lls_free_parse_result(lpr, CMD_PTR);
	if (lpr != cmdline_lpr)
		lls_free_parse_result(cmdline_lpr, CMD_PTR);
	if (ret < 0) {
		if (errctx)
			PARA_ERROR_LOG("%s\n", errctx);
		free(errctx);
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	}
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
