/*
 * Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file recv.h Receiver-related structures and exported symbols of recv_common.c. */

/**
 * Describes one instance of a receiver.
 */
struct receiver_node {
	/** Points to the corresponding receiver. */
	const struct receiver *receiver;
	/** Receiver-specific data. */
	void *private_data;
	/** The parsed command line options for this instance. */
	struct lls_parse_result *lpr;
	/** The task associated with this instance. */
	struct task *task;
	/** The receiver node is always the root of the buffer tree. */
	struct btr_node *btrn;
	/** Each receiver node maintains a buffer pool for the received data. */
	struct btr_pool *btrp;
	/**
	 * The file descriptor to receive the stream.
	 *
	 * The pre_select function of the receiver adds this file descriptor to
	 * the set of file descriptors which are watched for readability or
	 * writability, depending on the state of the connection (if any).
	 *
	 * If \a fd is readable, the post_select function of the receiver reads
	 * data from this fd into the buffer pool area of \a btrp.
	 *
	 * \sa \ref receiver.
	 */
	int fd;
};

/**
 * Describes one supported paraslash receiver.
 *
 * \sa http_recv.c, udp_recv.c
 */
struct receiver {
	/**
	 * The optional receiver init function.
	 *
	 * Performs any initialization needed before the receiver can be opened.
	 *
	 * \sa http_recv_init udp_recv_init.
	 */
	void (*init)(void);
	/**
	 * Open one instance of the receiver.
	 *
	 * This should allocate the output buffer of \a rn. and may also
	 * perform any other work necessary for retrieving the stream according
	 * to the configuration stored in the \a conf member of \a rn which is
	 * guaranteed to point to valid configuration data (as previously
	 * obtained from the config parser).
	 *
	 * \sa receiver_node::conf, receiver_node::buf.
	 */
	int (*open)(struct receiver_node *rn);
	/**
	 * Close this instance of the receiver.
	 *
	 * It should free all resources associated with given receiver node
	 * that were allocated during the corresponding open call.
	 *
	 * \sa receiver_node.
	 */
	void (*close)(struct receiver_node *rn);
	/**
	 * Add file descriptors to fd_sets and compute timeout for select(2).
	 *
	 * The pre_select function gets called from the driving application
	 * before entering its select loop. The receiver may use this hook to
	 * add any file descriptors to the sets of file descriptors given by \a
	 * s.
	 *
	 * \sa select(2), time.c struct task, struct sched.
	 */
	void (*pre_select)(struct sched *s, void *context);
	/**
	 * Evaluate the result from select().
	 *
	 * This hook gets called after the call to select(). It should check
	 * all file descriptors which were added to any of the fd sets during
	 * the previous call to pre_select. According to the result, it may
	 * then use any non-blocking I/O to establish a connection or to
	 * receive the audio data.
	 *
	 * \sa select(2), struct receiver.
	 */
	int (*post_select)(struct sched *s, void *context);

	/**
	 * Answer a buffer tree query.
	 *
	 * This optional function pointer is used for inter node communications
	 * of the buffer tree nodes. See \ref btr_command_handler for details.
	 */
	btr_command_handler execute;
};

#define RECV_CMD(_num) (lls_cmd(_num, recv_cmd_suite))

#define RECV_CMD_OPT_RESULT(_recv, _opt, _lpr) \
	(lls_opt_result(LSG_RECV_CMD_ ## _recv ## _OPT_ ## _opt, _lpr))
#define RECV_CMD_OPT_GIVEN(_recv, _opt, _lpr) \
	(lls_opt_given(RECV_CMD_OPT_RESULT(_recv, _opt, _lpr)))
#define RECV_CMD_OPT_STRING_VAL(_recv, _opt, _lpr) \
	(lls_string_val(0, RECV_CMD_OPT_RESULT(_recv, _opt, _lpr)))
#define RECV_CMD_OPT_UINT32_VAL(_recv, _opt, _lpr) \
	(lls_uint32_val(0, RECV_CMD_OPT_RESULT(_recv, _opt, _lpr)))
#define RECV_CMD_OPT_INT32_VAL(_recv, _opt, _lpr) \
	(lls_int32_val(0, RECV_CMD_OPT_RESULT(_recv, _opt, _lpr)))

/** Iterate over all available receivers. */
#define FOR_EACH_RECEIVER(i) for (i = 1; lls_cmd(i, recv_cmd_suite); i++)

void recv_init(void);
int check_receiver_arg(const char *ra, struct lls_parse_result **lprp);
void print_receiver_helps(bool detailed);
int generic_recv_pre_select(struct sched *s, struct receiver_node *rn);
